package pl.webnet.JaTrac;


class AreaLight extends Light implements XmlSettable
{
	Point x1, x2, x3, x4;
	Vec line1, line2;
	Vec normal;
	boolean visible = true;

	final static int SUBDIVISION = 5;
	final static double START_X = 1. / (SUBDIVISION * 2);
	final static double STEP_X = 1. / SUBDIVISION;
	final static double DARKEN = 1. / (SUBDIVISION * SUBDIVISION);

	public AreaLight()
	{
	}

	public AreaLight(Point x1, Point x2, Point x3, Point x4, Spectrum spectrum)
	{
		super();

		this.x1 = new Point(x1);
		this.x2 = new Point(x2);
		this.x3 = new Point(x3);
		this.x4 = new Point(x4);
		this.spectrum = spectrum;

		initialize();
	}

	public String[] getParamArray()
	{
		String params[] = {
			"visible", "visible", "",
			"p1", "x1", "",
			"p2", "x2", "",
			"p3", "x3", "",
			"p4", "x4", ""
		};

		return ParamHelper.mergeParams(super.getParamArray(), params);
	}

	public String[] getAttachArray() { return null; }

	public Object initialize()
	{
		line1 = new Vec(x2, x1);
		line2 = new Vec(x4, x1);
		normal = line1.cross(line2);
		normal.unit();

		origin = new Point((x1.x + x2.x + x3.x + x4.x) / 4,
												(x1.y + x2.y + x3.y + x4.y) / 4,
												(x1.z + x2.z + x3.z + x4.z) / 4);

		Level level = new Level();

		level.setParent(parent);

		// if light is visible, add an unshaded rectangle that will
		// imitate the light source object

		if(visible)
		{
			Rectangle r = new Rectangle(x1, x2, x3, x4);
			r.spectrum = spectrum.create(spectrum);
			//r.spectrum.scale(intensity);
			r.castShadows = false;
			r.unshaded = true;
			level.add(r);
		}

		level.add(this);

		return level;
	}

	public double shadow(RayTree rt, Point hitPoint, Vec lightVec, double len)
	{
		double shadow = 1.;
		double x, y;
		Ray lightRay;
		Primitive lastPrimitive = null;

		y = START_X;

		for(int i = 0; i < SUBDIVISION; i++)
		{
			x = START_X;

			for(int j = 0; j < SUBDIVISION; j++)
			{
				double tx = Math.random() / SUBDIVISION - START_X;
				double ty = Math.random() / SUBDIVISION - START_X;

				tx = x + tx;
				x += STEP_X;

				ty = y + ty;

				Vec target = new Vec(
					new Point(line1.x * tx + line2.x * ty + x1.x,
										line1.y * tx + line2.y * ty + x1.y,
										line1.z * tx + line2.z * ty + x1.z), hitPoint);

				lightRay = new Ray(hitPoint, target);


				// check if the object we hit with previous ray isn't
				// still blocking the light. If so, we can save some time
				// by not checking other primitives along the ray
				if(lastPrimitive != null)
				{
					if(lastPrimitive.intersect(lightRay) > 0)
					{
						if(((Intersection)lightRay.intersections.first()).t < len - Ray.EPSILON)
							{
								shadow -= DARKEN;
								continue;
							}
					}
					else
						lastPrimitive = null;
				}

				if(rt.iTester.intersectShadowfeeler(lightRay) != 0)
				{
					len = target.length();

					if(((Intersection)lightRay.intersections.first()).t < len - Ray.EPSILON)
						{
							lastPrimitive = ((Intersection)lightRay.intersections.first()).primitive;
							shadow -= DARKEN;
							continue;
						}
				}

				len = target.length() / 900;
				target.unit();

				shadow -= DARKEN * (Math.min(1, len * len + (1 - Math.abs(normal.dot(target)))));
			}
			
			y += STEP_X;
		}
		
		assert shadow >= 0 && shadow <= 1 : "Shadow value out of bounds" + shadow;
		return Math.max(0, shadow);
	}
}
