package pl.webnet.JaTrac;

class BoundingBox
{
	static final int X = 0;
	static final int Y = 1;
	static final int Z = 2;

	double x0, x1, y0, y1, z0, z1;

	public BoundingBox()
	{
		x0 = y0 = z0 = Double.MAX_VALUE;
		x1 = y1 = z1 = -Double.MAX_VALUE;
	}

	public BoundingBox(double x0, double x1, double y0, double y1, double z0, double z1)
	{
		set(x0, x1, y0, y1, z0, z1);
	}

	public void set(double x0, double x1, double y0, double y1, double z0, double z1)
	{
		this.x0 = x0;
		this.x1 = x1;
		this.y0 = y0;
		this.y1 = y1;
		this.z0 = z0;
		this.z1 = z1;
	}

	public double getMax(byte axis)
	{
		switch(axis)
		{
			case X: return(x1);
			case Y: return(y1);
			case Z: return(z1);
			default: throw new ArrayIndexOutOfBoundsException();
		}
	}

	public double getMin(byte axis)
	{
		switch(axis)
		{
			case X: return(x0);
			case Y: return(y0);
			case Z: return(z0);
			default: throw new ArrayIndexOutOfBoundsException();
		}
	}

	public void setMax(byte axis, double value)
	{
		switch(axis)
		{
			case X: x1 = value;
				break;
			case Y: y1 = value;
				break;
			case Z: z1 = value;
				break;
			default: throw new ArrayIndexOutOfBoundsException();
		}
	}

	public void setMin(byte axis, double value)
	{
		switch(axis)
		{
			case X: x0 = value;
				break;
			case Y: y0 = value;
				break;
			case Z: z0 = value;
				break;
			default: throw new ArrayIndexOutOfBoundsException();
		}
	}

	public void swallow(double x0, double x1, double y0, double y1, double z0, double z1)
	{
		if(x0 < this.x0)
			this.x0 = x0;
		if(x1 > this.x1)
			this.x1 = x1;

		if(y0 < this.y0)
			this.y0 = y0;
		if(y1 > this.y1)
			this.y1 = y1;

		if(z0 < this.z0)
			this.z0 = z0;
		if(z1 > this.z1)
			this.z1 = z1;
	}
	
	public void swallow(double x, double y, double z)
	{
		if(x < this.x0)
			this.x0 = x;
		if(x > this.x1)
			this.x1 = x;

		if(y < this.y0)
			this.y0 = y;
		if(y > this.y1)
			this.y1 = y;

		if(z < this.z0)
			this.z0 = z;
		if(z > this.z1)
			this.z1 = z;
	}

	public void swallow(BoundingBox bx)
	{
		swallow(bx.x0, bx.x1, bx.y0, bx.y1, bx.z0, bx.z1);
	}

	public boolean contains(Point p)
	{
		return(p.x >= x0 && p.x <= x1 &&
				p.y >= y0 && p.y <= y1 &&
				p.z >= z0 && p.z <= z1);
	}

	public double spanX()
	{
		return(x1 - x0);
	}

	public double spanY()
	{
		return(y1 - y0);
	}

	public double spanZ()
	{
		return(z1 - z0);
	}

	public double intersect(Ray ray)
	{
		double tNear = -Double.MAX_VALUE;
		double tFar = Double.MAX_VALUE;
		double t1, t2;

		if(ray.x == 0)
		{
			if(ray.origin.x < x0 || ray.origin.x > x1)
				return -1;
		}
		else
		{
			t1 = (x0 - ray.origin.x) / ray.x;
			t2 = (x1 - ray.origin.x) / ray.x;

			if(t1 > t2)
			{
				double t = t2;
				t2 = t1;
				t1 = t;
			}

			tNear = t1;
			tFar = t2;

			if(tFar < 0)
				return -1;
		}

		if(ray.y == 0)
		{
			if(ray.origin.y < y0 || ray.origin.y > y1)
				return -1;
		}
		else
		{
			t1 = (y0 - ray.origin.y) / ray.y;
			t2 = (y1 - ray.origin.y) / ray.y;

			if(t1 > t2)
			{
				double t = t2;
				t2 = t1;
				t1 = t;
			}

			if(t1 > tNear)
				tNear = t1;

			if(t2 < tFar)
				tFar = t2;

			if(tNear > tFar || tFar < 0)
				return -1;
		}

		if(ray.z == 0)
		{
			if(ray.origin.z < z0 || ray.origin.z > z1)
				return -1;
		}
		else
		{
			t1 = (z0 - ray.origin.z) / ray.z;
			t2 = (z1 - ray.origin.z) / ray.z;

			if(t1 > t2)
			{
				double t = t2;
				t2 = t1;
				t1 = t;
			}

			if(t1 > tNear)
				tNear = t1;

			if(t2 < tFar)
				tFar = t2;

			if(tNear > tFar || tFar < 0)
				return -1;
		}

		return tNear;
	}

	public String toString()
	{
		return(getClass().getName() + '@' + Integer.toHexString(hashCode()) + ": " + x0 + ", " + x1 + ", " + y0 + ", " + y1 + ", " + z0 + ", " + z1);
	}

}
