package pl.webnet.JaTrac;

class LineLight extends Light implements XmlSettable
{
	Point end;
	Vec lightLine;
	double raysPerDegree = 0.2;

	public LineLight()
	{
	}

	public LineLight(Point origin, Point end, Spectrum spectrum)
	{
		this.origin = new Point(origin);
		this.end = new Point(end);
		this.spectrum = spectrum;
	}

	public String[] getParamArray()
	{
		String params[] = {
			"p1", "origin", "",
			"p2", "end", "",
			"precision", "raysPerDegree", "",
			"spectrum", "", "setSpectrum"
		};

		return ParamHelper.mergeParams(super.getParamArray(), params);
	}

	public String[] getAttachArray() { return null; }
	public Object initialize()
	{
		lightLine = new Vec(end, origin);
		return this;
	}

	public double shadow(RayTree rt, Point hitPoint, Vec lightVec, double len)
	{
		double shadow = 1.;
		Ray lightRay;

		Vec l1 = new Vec(origin, hitPoint);
		Vec l2 = new Vec(end, hitPoint);
		l1.unit();
		l2.unit();
		double dot = l1.dot(l2);
		Vec line = new Vec(l2.x - l1.x, l2.y - l1.y, l2.z - l1.z);

		Point p = new Point(hitPoint.x + l1.x, hitPoint.y + l1.y, hitPoint.z + l1.z);

		int subdivision = (int)(Math.toDegrees(Math.acos(dot)) * raysPerDegree + 0.5);
		//int subdivision = 3;

//		return(SUBDIVISION / 180.);

		double startX = 1. / (subdivision * 2);
		double stepX = 1. / subdivision;
		double x = startX;

		for(int i = 0; i < subdivision; i++)
		{
			double t = Math.random() / subdivision - startX;
			t = x + t;
			x += stepX;

			Vec target = new Vec(new Point(line.x * t + p.x,
				line.y * t + p.y,
				line.z * t + p.z), hitPoint);

			lightRay = new Ray(hitPoint, target);

			// calculate distance to actual point on the light.
			double distance = ((origin.y - hitPoint.y) * lightLine.x + (hitPoint.x - origin.x) * lightLine.y) / (lightLine.x * lightRay.y - lightLine.y * lightRay.x);

			if(rt.iTester.intersectShadowfeeler(lightRay) != 0)
			{
//				target.unit();

				if(((Intersection)lightRay.intersections.first()).t < distance - Ray.EPSILON)
						shadow -= stepX;
			}
		}
		
		return Math.max(0., shadow);
	}
}
