package pl.webnet.JaTrac;

import java.awt.*;
import java.util.*;

/**
 * This class describes a generic primitive - ancestor of all
 * object types that can be intersected by a ray.
 * @author: Miloslaw Smyk (thorgal@amiga.com.pl)
 * @version: $Id: Primitive.java,v 1.5 2001/06/03 01:42:38 thorgal Exp $
 */


abstract class Primitive extends TreeNode implements XmlSettable
{
	Spectrum spectrum = null;
	Mapping mapping = null;
	boolean shadows = true;
	boolean castShadows = true;
	boolean reflected = true;
	boolean unshaded = false;
	Intersection lastIntersection;
	Point lastHitPoint;
	long mailbox;
	String name = null;

	public Primitive()
	{
		super();
	}
/**
 * Check if the ray intersects this primitive and compute intersection points
 * as a value of <code>t</code> along the ray.
 * @param		ray the ray to test against
 * @return	the number of intersections
 */
	abstract public int intersect(Ray ray);

/**
 * Compute normal vector at given point on object's surface.
 * @param		p point (in world coordinates) on object's surface
 * @return	normal vector
 */
	abstract public Vec getNormal(Point p);
	
/**
 * Compute tangent vector at given point on object's surface.
 * Used for bump mapping.
 * @param		n surace's normal vector at this point. If null, will be computed.
 * @param		p point (in world coordinates) on object's surface
 * @return	normal vector
 */
	abstract public Vec getTangent(Point p, Vec normal);

	abstract public BoundingBox getBoundingBox(BoundingBox bx);
	
	public Surface getSurface()
	{
		if(mapping != null)
			return mapping.surface;
		else
			return parent.averageSurface;
	}

	public Mapping getMapping() { return mapping; }

	public Spectrum getSpectrum() { return spectrum; }

	public void setSpectrum(Spectrum s)	{ spectrum = s;	}

	public void setName(String name) { this.name = name; }

	public String[] getParamArray()
	{
		String params[] = {
			"name", "name", "",
			"shadows", "shadows", "",
			"castshadows", "castShadows", "",
			"reflected", "reflected", "",
			"unshaded", "unshaded", "",
			"mapping", "mapping", "",
			"spectrum", "spectrum", ""
		};

		return params;
	}

	public String[] getAttachArray()
	{
		return null;
	}

	public Object initialize() { return this; }
	
}
