package pl.webnet.JaTrac;

import java.util.*;
import java.awt.*;

class RayTree
{
	Scene scene;
	IntersectionTester iTester;
	final Double noShadowConst = new Double(1);
	// BoundingVolume bound;

	int maxRecurseLevel;

	public RayTree(Scene s, int maxLevel)
	{
		scene = s;
		iTester = scene.iTester;

		maxRecurseLevel = maxLevel;
	}

	public Ray trace(int level, double weight, Ray ray)
	{
		Primitive prim;
		Point hitPoint;
		Vec normalVec;
		Intersection intersection;
		
		//if(intersect(ray, bound, 0) != 0)
		if(iTester.intersectRay(ray) != 0)
		{
			intersection = (Intersection)ray.intersections.first();

			hitPoint = ray.rayPoint(intersection.t);
			normalVec = intersection.primitive.getNormal(hitPoint);

			if(ray.dot(normalVec) > 0.)
				normalVec.negate();

			ray.spectrum = shade(level, weight, hitPoint, normalVec, ray, intersection);
		}
		else
			if(scene.camera.mediumSpectrum == null)
				ray.spectrum = scene.spectrum.create(0, 0, 0);
			else
				ray.spectrum = scene.camera.mediumSpectrum;

		/*
			if(level == 0)
				//ray.spectrum = scene.spectrum.create(88, 35, 149);
				ray.spectrum = scene.camera.mediumSpectrum;
			else
				//ray.spectrum = scene.spectrum.create(128, 0, 0);
				ray.spectrum = scene.camera.mediumSpectrum;
		*/

		return ray;
	}

	public Spectrum processTextures(Primitive prim, Point hitPoint, Vec normalVec)
	{
		Mapping mn;
		Spectrum primSpectrum = prim.getSpectrum();

		// if primitive's spectrum is not defined, use a global one
		if(primSpectrum == null)
			primSpectrum = scene.spectrum;

		Vec tangent = prim.getTangent(hitPoint, normalVec);

		// if primitive has its own private mapping, use it
		if(((mn = prim.getMapping()) != null))
		{
			if(mn.surface.texture != null)
			{
				Point uv = mn.getUV(hitPoint, normalVec);

				int prop = mn.surface.texture.getProperties();

				if((prop & Texture.TEXTURE_NORMALVEC) != 0)
					normalVec = mn.surface.texture.updateNormal(normalVec, tangent, uv);

				if((prop & Texture.TEXTURE_SPECTRUM) != 0)
					primSpectrum = mn.surface.texture.updateSpectrum(primSpectrum, uv);
			}
		}
		else
		{
			// otherwise:
			// go over all mappings affecting this level and calculate
			// average texture properties like colors and normal vector
			Spectrum textureSpectrum = scene.spectrum.create(Spectrum.black);
			double textureTotalWeight = 0;

			for(Iterator i = prim.parent.affectingMappings.iterator(); i.hasNext(); )
			{
				mn = (Mapping)i.next();
		
				Point uv = mn.getUV(hitPoint, normalVec);

				if(mn.surface.texture != null)
				{
					int prop = mn.surface.texture.getProperties();

					// FIXME: needs to average normals, not replace them!
					if((prop & Texture.TEXTURE_NORMALVEC) != 0)
						normalVec = mn.surface.texture.updateNormal(normalVec, tangent, uv);

					if((prop & Texture.TEXTURE_SPECTRUM) != 0)
					{
					//	mn.surface.texture.updateSpectrum(primSpectrum, uv).scale(mn.surface.weight);
						//textureSpectrum.add(mn.surface.texture.updateSpectrum(primSpectrum, uv));
						textureSpectrum.add(mn.surface.texture.updateSpectrum(primSpectrum, uv).scale(mn.surface.weight));
						textureTotalWeight += mn.surface.weight;
					}
				}
			}

			if(textureTotalWeight > 0)
					primSpectrum = textureSpectrum.scale(1. / textureTotalWeight);
		}

		return primSpectrum;
	}

	public Spectrum shade(int level, double weight, Point hitPoint, Vec normalVec, Ray ray, Intersection intersection)
	{
		Primitive prim = intersection.primitive;
		Spectrum diffPrimSpectrum, specPrimSpectrum;
		Spectrum primSpectrum;
		Spectrum reflSpectrum;
		Spectrum transSpectrum;
		Spectrum diffSpectrum = scene.spectrum.create(Spectrum.black);
		Spectrum specSpectrum = scene.spectrum.create(Spectrum.black);

		double LRdot;
		double lightAngle;
		double shadowCache;
		Surface surface = prim.getSurface();

		// first, compute modifications to surface color and normal vector
		// possibly introduced by color and bump-mapping
		primSpectrum = processTextures(prim, hitPoint, normalVec);

		// we do not perform lighting calculations if primitive has unshaded
		// flag set.
		if(prim.unshaded)
			return(primSpectrum);

		diffPrimSpectrum = primSpectrum.create(primSpectrum);
		diffPrimSpectrum.scale(surface.diffuseCoeff);

		specPrimSpectrum = primSpectrum.create(primSpectrum);
		specPrimSpectrum.scale(surface.specularCoeff);

		// create unit specular vector
		Ray specularRay = generateSpecularRay(hitPoint, normalVec, ray);
		//specularRay.unit();

		// iterate over all light sources gathering diffuse light and highlights

		for(Iterator i = scene.lights.iterator(); i.hasNext(); )
		{
			if(prim.shadows == false || scene.camera.screen.shadows == false)
				shadowCache = 1;
			else
				shadowCache = -1;

			Light light = (Light)i.next();

			Vec lightVec = new Vec(light.getOrigin(), hitPoint);
			double lightVecLen = lightVec.length();

			lightVec.unit();

			// currently area lights are handled completely differently
			// (ok, it's a hack). this is going to change.
			if(light instanceof AreaLight)
			{
					if(shadowCache != -1 || (shadowCache = light.shadow(this, hitPoint, lightVec, lightVecLen)) > 0.)
						// multiply and accumulate
						diffSpectrum.MAC(shadowCache * light.intensity, light.getSpectrum());
			}
			else
				if((lightAngle = normalVec.dot(lightVec)) > 0.)
				{
					if(shadowCache != -1 || (shadowCache = light.shadow(this, hitPoint, lightVec, lightVecLen)) > 0.)
					{
						// multiply and accumulate
						diffSpectrum.MAC(lightAngle * shadowCache * light.intensity, light.getSpectrum());
//						diffPrimSpectrum = primSpectrum.create(Spectrum.white);
	//					diffPrimSpectrum.scale(shadowCache);
		//				return diffPrimSpectrum;
					}
			}

			LRdot = specularRay.dot(lightVec);

			if(LRdot > 0.)
			{
				// if no shadowfeeler was shot to the light yet, do it now
				// otherwise, reuse
				if(shadowCache != -1 || (shadowCache = light.shadow(this, hitPoint, lightVec, lightVecLen)) > 0.)
				{
					double specular = shadowCache * Math.pow(specularRay.dot(lightVec), surface.specularSizeCoeff);
					specSpectrum.brighten(specular);
				}
			}
		}

		if(scene.camera.ambientSpectrum != null)
			diffSpectrum.add(scene.camera.ambientSpectrum);

		// calculate the actual fraction of diffuse and specular spectrum reflected
		// from the surface
		diffSpectrum.interact(diffPrimSpectrum);
		specSpectrum.interact(specPrimSpectrum);

		if(level + 1 <= maxRecurseLevel)
		{
			if(surface.reflectionCoeff * weight > .05)
			{
				ray.reflectedRay = trace(level + 1, surface.reflectionCoeff * weight, specularRay);

				reflSpectrum = ray.reflectedRay.spectrum.create(ray.reflectedRay.spectrum);
				reflSpectrum.interact(surface.reflectionCoeff, primSpectrum);
				diffSpectrum.add(reflSpectrum);
			}

			if(surface.transmissionCoeff * weight > 0.05)
			{
				Ray tr = generateTransmittedRay(hitPoint, normalVec, ray);

				if(tr == null)
					tr = specularRay;

				ray.transmittedRay = trace(level + 1, surface.transmissionCoeff * weight, tr);

				transSpectrum = ray.transmittedRay.spectrum.create(ray.transmittedRay.spectrum);
				transSpectrum.interact(surface.transmissionCoeff, primSpectrum);
				diffSpectrum.add(transSpectrum);
			}

		}
			
		diffSpectrum.add(specSpectrum);
//		Spectrum medium = scene.spectrum.create(120, 35, 120);
	//	double diminish = Math.max(0, 1 - intersection.t / 40);

		//diffSpectrum.scale(diminish);
	//	medium.scale(1 - diminish);
//		diffSpectrum.add(medium);
		return diffSpectrum;
	}

	public Ray generateSpecularRay(Point hitPoint, Vec normalVec, Ray rayDir)
	{
		Ray r = new Ray(hitPoint, normalVec.addScalarMultiple(-2. * normalVec.dot(rayDir), rayDir));
		r.generation = rayDir.generation + 1;

		return r;
	}

	public Ray generateTransmittedRay(Point hitPoint, Vec normalVec, Ray incidentRay)
	{
		double ri1 = 1.0, ri2 = 1.0;
	/*	
		Surface m[] = new Surface[2];

		for(Iterator i =  (Intersection)incidentRay.intersections.iterator(); i.hasNext(); )
		{
			Intersection is = (Intersection)i.next();
			m[idx++] = is.surface;

			if(idx == 2)
				break;
		}

		if(m[0]
		*/

		Surface s = ((Intersection)incidentRay.intersections.first()).surface;
		
		if(s != null)
			if(((Intersection)incidentRay.intersections.first()).entering == Intersection.ENTERING)
				ri2 = s.refractionIndex;
			else
				ri1 = s.refractionIndex;

		double eta = ri1 / ri2;
		double c1 = -incidentRay.dot(normalVec);
		double cs2 = 1. - eta * eta * (1. - c1 * c1);

		if(cs2 < 0.)
			return null;

		return (new Ray(hitPoint, incidentRay.linearCombination(eta, eta * c1 - Math.sqrt(cs2), normalVec)));
	}
}
