package pl.webnet.JaTrac;

import java.util.*;

/**
 * This class contains a complete scene with lists of all primitives,
 * lights and mappings.
 * @author: Miloslaw Smyk (thorgal@amiga.com.pl)
 * @version: $Id: Scene.java,v 1.5 2001/06/03 01:42:38 thorgal Exp $
 */

class Scene implements XmlSettable
{
	Camera camera;
	Spectrum spectrum;
	Level rootNode;
	static Vector lights;

	// use naive intersection tester by default. hopefully scene
	// definition will replace it with something faster.
	IntersectionTester iTester = new NaiveTester();


	public Scene()
	{
		lights = new Vector();
		add(new Level());
	}

/**
 * This Iterator is used to iterate through all primitives in the scene.
 * It's somewhat slower than direct recursion, so Intersectors shouldn't use
 * it, but for other uses it is very convenient.
 */
	public Iterator iterator()
	{
		return new Iterator()
		{
			Stack levelStack = new Stack();
			Iterator i = rootNode.children.iterator();
			Object next = null;

			public void remove()
			{
				throw new UnsupportedOperationException("Scene Iterator");
			}
			
			public boolean hasNext()
			{
				try
				{
					if(next == null)
						next = next();

					return true;
				} catch(NoSuchElementException e)
				{
					return false;
				}
			}

			public Object next()
			{
				if(next != null)
				{
					Object current = next;
					next = null;
					return current;
				}

				if(i.hasNext())
				{
					TreeNode tn = (TreeNode)i.next();

					if(tn instanceof Level)
					{
						levelStack.push(i);
						i = ((Level)tn).children.iterator();
						return next();
					}

					return tn;
				}
				else
				{
					try
					{
						i = (Iterator)levelStack.pop();
						return next();
					} catch (EmptyStackException e)
					{
						throw new NoSuchElementException("Scene Iterator");
					}
				}
			}
		};
	}
	
/**
 * Add new light to the scene.
 * @param		light the light to be put on the scene
 */
	static public void add(Light light)
	{
		lights.add(light);
	}

	public void add(TreeNode tn)
	{
		rootNode.add(tn);
	}

	public void add(Level level)
	{
		rootNode = level;
		TreeNode.setRoot(rootNode);
		rootNode.setParent(null);
	}

	public void set(IntersectionTester it)
	{
		iTester = it;
		System.out.println("Using intersection tester: " + it);
	}

	public String[] getParamArray()
	{
		String params[] = {
		};

		return params;
	}

	public String[] getAttachArray()
	{
		String params[] = {
			"Camera", "setCamera", "",
			"Spectrum", "setSpectrum", "",
			"Light", "add", "",
			"Level", "add", "",
			"IntersectionTester", "set", ""
		};

		return params;
	}

	public Object initialize() { return this; }

	public void setCamera(Camera cam) { camera = cam; }
	public void setSpectrum(Spectrum s) { spectrum = s; }

/**
 * Go through levels of the hierarchy and calculate what Surfaces
 * affect each level. Store this average surface in each level
 * instance for quicker lookup of surface parameters during rendering.
 */
	public void preprocessSurfaces()
	{
		System.out.println("Preprocessing surfaces...");

		preprocessSurfaces(rootNode, new Vector());
	}

/**
 * Go through levels of the hierarchy and calculate what Surfaces
 * affect each level. Store this average surface in each level
 * instance for quicker lookup of surface parameters during rendering.
 * @param		level top node of the tree to process
 * @param		collected list of surfaces collected as yet (for recursive use)
 */
	private void preprocessSurfaces(Level level, Vector collected)
	{
		// clear the list of surfaces affecting this level and initialize
		// it with surfaces from higher levels
		level.affectingMappings = new Vector(collected);

		// go through all mappings on this level and include them on our list
		for(Iterator i = level.mappings.iterator(); i.hasNext(); )
		{
			Mapping mn = (Mapping)i.next();

			level.affectingMappings.add(mn);
		}

		level.averageSurface = Surface.averageSurfaces(level.affectingMappings);

		// finally go through all children and initialize sublevels
		for(Iterator i = level.children.iterator(); i.hasNext(); )
		{
			TreeNode tn = (TreeNode)i.next();
			
			if(tn instanceof Level)
				preprocessSurfaces((Level)tn, level.affectingMappings);
		}
	}

	public String toString()
	{
		return(ParamHelper.dumpAttrs(this));
	}

}
