package pl.webnet.JaTrac;

import java.awt.*;

/**
 * Sphere implements spherical primitive.
 * @author: Miloslaw Smyk (thorgal@amiga.com.pl)
 * @version: $Id: Sphere.java,v 1.4 2001/06/03 01:42:38 thorgal Exp $
 */


class Sphere extends Primitive implements XmlSettable
{
	Point center;
	double radius, radiusSquared;
	static Vec unit = new Vec(0., 1., 0.);
	static int count = 0;


/**
 * Construct a sphere.
 * @param		center sphere's center in world coordinates
 * @param		radius sphere's radius
 */
	public Sphere(Point center, double radius)
	{
		super();

		this.center = new Point(center);
		this.radius = radius;
		radiusSquared = radius * radius;
	}

	public Sphere()
	{
		super();
	}

	public String[] getParamArray()
	{
		final String params[] = {
			"origin", "center", "",
			"radius", "radius", "",
		};

		return ParamHelper.mergeParams(super.getParamArray(), params);
	}

	public String[] getAttachArray() { return null; }

	public Object initialize()
	{
		radiusSquared = radius * radius;

		return this;
	}

	public void setRadius(String rad)
	{
		double r = ParamHelper.extractDouble(rad);
		radius = r;
	}

/**
 * Check if given ray intersects the sphere. Uses geometric solution from
 * <em>An Introduction to Raytracing</em> by A. S. Glassner et all.
 * @param		ray the ray to test against
 * @return	the number of intersections
 */
	public int intersect(Ray ray)
	{
		count++;
		Vec oc = new Vec(center, ray.origin);

		double b = oc.dot(ray);
		double disc = b * b - oc.dot(oc) + radiusSquared;
		int nroots = 0;
		double t1, t2;

		if(disc <= 0.)
			return 0;

		disc = Math.sqrt(disc);
		t2 = b + disc;

		if(t2 <= Ray.EPSILON)
			return 0;

		t1 = b - disc;

		if(t1 > Ray.EPSILON)
		{
			ray.intersections.add(new Intersection(t1, this, Intersection.ENTERING, getSurface()));
			nroots = 2;
		}
		else
			nroots = 1;

			ray.intersections.add(new Intersection(t2, this, Intersection.EXITING, getSurface()));

			return nroots;
	}

	public Vec getTangent(Point p, Vec n)
	{
		Vec tangent;
		double len = unit.dot(n);

		if(n == null)
			n = new Vec(p, center);

		tangent = new Vec(- len * n.x, 1 - len * n.y, - len * n.z);

		tangent.unit();

		return tangent;
	}

/**
 * Return unit normal in given point.
 * @param		p point to calculate normal at
 * @return	unit normal vector.
 */
	public Vec getNormal(Point p)
	{
		Vec normal = new Vec(p, center);

		normal.unit();

		return normal;		
	}

	public BoundingBox getBoundingBox(BoundingBox bx)
	{
		bx.set(center.x - radius, center.x + radius,
						center.y - radius, center.y + radius,
						center.z - radius, center.z + radius);
		return bx;
	}

	public String toString()
	{
		return(getClass().getName() + '@' + Integer.toHexString(hashCode()) + ": c:" + center + ", r: " + radius);
	}

}
